"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.main = void 0;
const node_path_1 = __importDefault(require("node:path"));
const typescript_1 = __importDefault(require("typescript"));
const config_1 = require("./util/config");
const fs_1 = require("./util/fs");
const path_1 = require("./util/path");
const project_1 = require("./util/project");
const runner_1 = require("./runner");
const main = async (options) => {
    const { cwd, workingDir, configFilePath = 'knip.json', tsConfigFilePath, include, exclude, ignore, gitignore, isDev, isShowProgress, jsDoc, } = options;
    const localConfigurationPath = configFilePath && (await (0, fs_1.findFile)(workingDir, configFilePath));
    const manifestPath = await (0, fs_1.findFile)(workingDir, 'package.json');
    const localConfiguration = localConfigurationPath && require(localConfigurationPath);
    const manifest = manifestPath && require(manifestPath);
    if (!localConfigurationPath && !manifest.knip) {
        const location = workingDir === cwd ? 'current directory' : `${node_path_1.default.relative(cwd, workingDir)} or up.`;
        throw new Error(`Unable to find ${configFilePath} or package.json#knip in ${location}`);
    }
    const dir = node_path_1.default.relative(cwd, workingDir);
    const resolvedConfig = (0, config_1.resolveConfig)(manifest.knip ?? localConfiguration, { workingDir: dir, isDev });
    if (!resolvedConfig) {
        throw new Error('Unable to find `entryFiles` and/or `projectFiles` in configuration.');
    }
    const report = (0, config_1.resolveIncludedIssueGroups)(include, exclude, resolvedConfig);
    let tsConfigPaths = [];
    const tsConfigPath = await (0, fs_1.findFile)(workingDir, tsConfigFilePath ?? 'tsconfig.json');
    if (tsConfigFilePath && !tsConfigPath) {
        throw new Error(`Unable to find ${tsConfigFilePath}`);
    }
    if (tsConfigPath) {
        const tsConfig = typescript_1.default.readConfigFile(tsConfigPath, typescript_1.default.sys.readFile);
        tsConfigPaths = tsConfig.config.compilerOptions?.paths
            ? Object.keys(tsConfig.config.compilerOptions.paths).map(p => p.replace(/\*/g, '**'))
            : [];
        if (tsConfig.error) {
            throw new Error(`An error occured when reading ${node_path_1.default.relative(cwd, tsConfigPath)}`);
        }
    }
    const projectOptions = tsConfigFilePath ? { tsConfigFilePath } : { compilerOptions: { allowJs: true } };
    const entryPaths = await (0, path_1.resolvePaths)({
        cwd,
        workingDir,
        patterns: resolvedConfig.entryFiles,
        ignore,
        gitignore,
    });
    const production = (0, project_1.createProject)({ projectOptions, paths: entryPaths });
    const entryFiles = production.getSourceFiles();
    production.resolveSourceFileDependencies();
    const productionFiles = production.getSourceFiles();
    const projectPaths = await (0, path_1.resolvePaths)({
        cwd,
        workingDir,
        patterns: resolvedConfig.projectFiles,
        ignore,
        gitignore,
    });
    const project = (0, project_1.createProject)({ projectOptions, paths: projectPaths });
    const projectFiles = project.getSourceFiles();
    const config = {
        workingDir,
        report,
        entryFiles,
        productionFiles,
        projectFiles,
        dependencies: Object.keys(manifest.dependencies ?? {}),
        devDependencies: Object.keys(manifest.devDependencies ?? {}),
        isDev: typeof resolvedConfig.dev === 'boolean' ? resolvedConfig.dev : isDev,
        tsConfigPaths,
        isShowProgress,
        jsDocOptions: {
            isReadPublicTag: jsDoc.includes('public'),
        },
    };
    const { issues, counters } = await (0, runner_1.findIssues)(config);
    return { report, issues, counters };
};
exports.main = main;
